import numpy as np
import traitlets
import vaex.ml.state
import logging

import vaex
import vaex.serialize
# vaex.set_log_level_debug()

logger_km = logging.getLogger('vaex.ml.kmeans')

def Matrix(type=traitlets.CFloat):
    return traitlets.List(traitlets.List(type()))


from numba import jit
#@jit('void(double[:], double[:], double[:])', nopython=True, nogil=True)
@jit(nopython=True, nogil=True, cache=True)
def distances_square(result, centroids, *blocks):
    """
    Function under test.
    """
    k = centroids.shape[0]
    dimensions = centroids.shape[1]
    N = result.shape[1]
    for i in range(k):
        for d in range(dimensions):
            for j in range(N):
                result[i,j] += (blocks[d][j] - centroids[i][d])**2

@jit(nopython=True, nogil=True, cache=True)
def centroid_stats(centroids, counts, sumpos, inertia, *blocks):
    #classes = np.argmin(distances_sq, axis=0)
    #distances_sq = np.choose(classes, distances_sq)
    #inertia = (distances_sq).sum()
    #inertia = 0
    N = blocks[0].shape[0]
    runs = sumpos.shape[0]
    clusters = sumpos.shape[1]
    dimensions = sumpos.shape[2]
    for j in range(N):
        for run in range(runs):
            #if done[run]: ## TODO: bug in numba? seems to crash if included
                best_distance = 1e100
                best_class = 10000
                for cluster in range(clusters):
                    distance = 0
                    for d in range(dimensions):
                        distance += (centroids[run,cluster,d] - blocks[d][j])**2
                    if distance < best_distance:
                        best_class = cluster
                        best_distance = distance
                cls = best_class #classes[j]
                inertia[run] += best_distance
                for d in range(dimensions):
                     sumpos[run,cls,d] += blocks[d][j]
                counts[run,cls] += 1
#    return inertia
        #for i in range(k):

@vaex.serialize.register
class KMeans(vaex.ml.state.HasState):
    prediction_label = traitlets.Unicode(default_value='prediction_kmeans')
    n_clusters = traitlets.CInt(default_value=2)
    init = traitlets.Union([Matrix(), traitlets.Unicode()], default_value='random')
    max_iter = traitlets.CInt(default_value=300)
    inertia = traitlets.CFloat(default_value=None, allow_none=True)
    random_state = traitlets.CInt(default_value=None, allow_none=True)
    cluster_centers = traitlets.List(traitlets.List(traitlets.CFloat()))
    features = traitlets.List(traitlets.Unicode())
    verbose = traitlets.CBool(default_value=False)
    n_init = traitlets.CInt(default_value=1)

    def __call__(self, *blocks):
        return self._calculate_classes(*blocks)

    def _calculate_distances_squared(self, *blocks):
        N = len(blocks[0]) # they are all the same length
        centroids = np.array(self.cluster_centers)
        k = centroids.shape[0]
        dimensions = centroids.shape[1]
        distances_sq = np.zeros((k, N))
        if 1:
            distances_square(distances_sq, centroids, *blocks)
        else:
            for d in range(dimensions):
                for i in range(k):
                    distances_sq[i] += (blocks[d] - centroids[i][d])**2
        return distances_sq

    def _calculate_classes(self, *blocks):
        distances_sq = self._calculate_distances_squared(*blocks)
        classes = np.argmin(distances_sq, axis=0)
        return classes

    def generate_cluster_centers_random(self, dataset, rng):
        indices = rng.randint(0, len(dataset), self.n_clusters)
        return [[dataset.evaluate(feature, i1=i, i2=i+1)[0] for feature in self.features] for i in indices]

    def transform(self, dataset):
        copy = dataset.copy()
        lazy_function = copy.add_function('kmean_predict_function', self, unique=True)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_label, expression, unique=False)
        return copy

    def fit(self, dataset):
        if self.init == 'random':
            rng = np.random.RandomState(self.random_state)
            self.run_cluster_centers = [self.generate_cluster_centers_random(dataset, rng) for k in range(self.n_init)]
        else:
            if self.n_init > 1:
                print("WARNING: n_init > 1 , but init given, only doing one run")
            self.run_cluster_centers = [self.init]
        done = [False] * len(self.run_cluster_centers)
        alldone = False
        first = True
        previous_inertias = None
        iteration = 0
        inertias_list = []
        while not alldone:
            new_centers, inertias = self._find_centers_and_inertias(dataset, done)
            if self.verbose:
                inertia_msges = [('--' if d else '{: 3}'.format(k)) for k, d in zip(inertias, done)]
                if len(inertias) == 1:
                    inertia_msg = inertia_msges[0]
                else:
                    inertia_msg = " | ".join(inertia_msges)
                print('Iteration {: 4}, inertia {}'.format(iteration, inertia_msg))
            if not first:  # we can only to a check after the second iteration
                done = self._is_done(previous_inertias, inertias)
                alldone = np.all(done)
            else:
                first = False
            iteration += 1
            if (iteration >= self.max_iter):
                logger_km.debug('reached max iterations: %s', self.max_iter)
                alldone = True
            previous_inertias = inertias
            inertias_list.append(inertias)
            self.run_cluster_centers = new_centers
        best_run = np.argmin(inertias)
        self.inertia = inertias[best_run]
        self.inertias = np.array(inertias_list)[:,best_run]
        self.cluster_centers = new_centers[best_run]

    def _is_done(self, inertias1, inertias2):
        diffs = [(inertia1 - inertia2) for inertia1, inertia2 in zip(inertias1, inertias2)]
        return [(diff < 1e-3) for diff in diffs]

    def _find_centers_and_inertias(self, dataset, done):
        done = np.array(done, dtype=np.int8)
        centroids = np.array(self.run_cluster_centers)
        runs = centroids.shape[0]
        clusters = centroids.shape[1]
        dimensions = centroids.shape[2]
        #print("k =", k)
        assert dimensions == len(self.features), "nr of dimensions for centroid should equal nr of features"

        def map(*blocks): # this will be called with a chunk of the data
            sumpos = np.zeros((runs, clusters, dimensions))
            counts = np.zeros((runs, clusters))
            inertia = np.zeros((runs))
            if 1:
                centroid_stats(centroids, counts, sumpos, inertia,  *blocks)
            else:
                # this is the pure python code
                # although not made for multiple runs yet
                distances_sq = self._calculate_distances_squared(*blocks)
                classes = np.argmin(distances_sq, axis=0)
                distances_sq = np.choose(classes, distances_sq)
                inertia = (distances_sq).sum()
                for i in range(k):
                    mask = classes == i
                    counts[i] = np.sum(mask)
                    for d in range(dimensions):
                        sumpos[i,d] += blocks[d][mask].sum()
            return sumpos, counts, inertia
        def reduce(x, y):
            sumpos1, counts1, inertia1 = x
            sumpos2, counts2, inertia2 = y
            return sumpos1 + sumpos2, counts1 + counts2, inertia1 + inertia2
        sumpos, counts, inertia = dataset.map_reduce(map, reduce, self.features)
        means = (sumpos/counts[:,:,np.newaxis])
        return means.tolist(), inertia



import functools
import vaex.dataset
from vaex.serialize import register
import numpy as np
from . import generate
from .state import HasState
from traitlets import Dict, Unicode, List, Int, CFloat, CBool


def dot_product(a, b):
    products = ['%s * %s' % (ai, bi) for ai, bi in zip(a, b)]
    return ' + '.join(products)


@register
class StateTransfer(HasState):
    state = Dict()

    def transform(self, dataset):
        copy = dataset.copy()
        self.state = dict(self.state, active_range=[copy._index_start, copy._index_end])
        copy.state_set(self.state)
        return copy


@register
@generate.register
class PCA(HasState):

    features = List(Unicode())
    n_components = Int(default_value=2)
    eigen_vectors = List(List(CFloat)).tag(output=True)
    eigen_values = List(CFloat).tag(output=True)
    prefix = Unicode(default_value="PCA_")
    means = List(CFloat).tag(output=True)

    def fit(self, dataset, column_names=None, progress=False):
        assert self.n_components <= len(self.features), 'cannot have more components than features'
        C = dataset.cov(self.features, progress=progress)
        eigen_values, eigen_vectors = np.linalg.eigh(C)
        indices = np.argsort(eigen_values)[::-1]
        self.means = dataset.mean(self.features, progress=progress).tolist()
        self.eigen_vectors = eigen_vectors[:, indices].tolist()
        self.eigen_values = eigen_values[indices].tolist()

    def transform(self, dataset, n_components=None):
        n_components = n_components or self.n_components
        copy = dataset.copy()
        name_prefix_offset = 0
        eigen_vectors = np.array(self.eigen_vectors)
        while self.prefix + str(name_prefix_offset) in copy.get_column_names(virtual=True, strings=True):
            name_prefix_offset += 1

        expressions = [copy[feature]-mean for feature, mean in zip(self.features, self.means)]
        for i in range(n_components):
            v = eigen_vectors[:, i]
            expr = dot_product(expressions, v)
            name = self.prefix + str(i + name_prefix_offset)
            copy[name] = expr
        return copy


@register
@generate.register
class StandardScaler(HasState):

    features = List(Unicode())
    mean = List(CFloat).tag(output=True)
    std = List(CFloat).tag(output=True)
    prefix = Unicode(default_value="standard_scaled_")
    with_mean = CBool(default_value=False)
    with_std = CBool(default_value=False)

    def fit(self, dataset):

        mean = dataset.mean(self.features, delay=True)
        std = dataset.std(self.features, delay=True)

        @vaex.delayed
        def assign(mean, std):
            self.mean = mean.tolist()
            self.std = std.tolist()

        assign(mean, std)
        dataset.executor.execute()

    def transform(self, dataset):
        copy = dataset.copy()
        for i in range(len(self.features)):
            name = self.prefix+self.features[i]
            expression = copy[self.features[i]]
            if self.with_mean:
                expression = expression - self.mean[i]
            if self.with_std:
                expression = expression / self.std[i]
            copy[name] = expression
        return copy


@register
@generate.register
class MinMaxScaler(HasState):

    features = List(Unicode())
    prefix = Unicode(default_value="standard_scaled_")
    fmin = List(CFloat).tag(output=True)
    fmax = List(CFloat).tag(output=True)
    feature_range = List(CFloat).tag(output=True)
    prefix = Unicode(default_value="minmax_scaled_")

    def fit(self, dataset):
        assert len(self.feature_range) == 2, 'feature_range must have 2 elements only'
        minmax = dataset.minmax(self.features)
        self.fmin = minmax[:, 0].tolist()
        self.fmax = minmax[:, 1].tolist()

    def transform(self, dataset):
        copy = dataset.copy()

        for i in range(len(self.features)):
            name = self.prefix + self.features[i]
            a = self.feature_range[0]
            b = self.feature_range[1]
            expr = copy[self.features[i]]
            expr = (b-a)*(expr-self.fmin[i])/(self.fmax[i]-self.fmin[i]) + a
            copy[name] = expr
        return copy





import vaex.ml.transformations
import traitlets

def pca(self, features, n_components=2, prefix='PCA_'):
    obj = vaex.ml.transformations.PCA(n_components=n_components, prefix=prefix, features=features)
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(pca=pca)

def __init__(self, n_components=2, prefix='PCA_', features=traitlets.Undefined):
    super(vaex.ml.transformations.PCA, self).__init__(n_components=n_components, prefix=prefix, features=features)

vaex.ml.transformations.PCA.__init__ = __init__
del __init__
    



import numpy as np

#pythran export distances(float64 [][][], float64 [][], float64 [][][], float64 [], float64 [][])
def distances(centroids, counts, sumpos, inertia, done, blocks):
    N = blocks[0].shape[0]
    runs = sumpos.shape[0]
    clusters = sumpos.shape[1]
    dimensions = sumpos.shape[2]
    for j in range(N):
        for run in range(runs):
                best_distance = 1e100
                best_class = 10000
                for cluster in range(clusters):
                    distance = 0
                    for d in range(dimensions):
                        distance += (centroids[run,cluster,d] - blocks[d,j])**2
                    if distance < best_distance:
                        best_class = cluster
                        best_distance = distance
                cls = best_class #classes[j]
                inertia[run] += best_distance
                for d in range(dimensions):
                     sumpos[run,cls,d] += blocks[d,j]
                counts[run,cls] += 1



import vaex.utils
from traitlets import *

class HasState(HasTraits):

    @classmethod
    def state_from(cls, state):
        obj = cls()
        obj.state_set(state)
        return obj

    def state_get(self):
        return {name: getattr(self, name) for name in self.trait_names()}

    def state_set(self, state):
        for name in self.trait_names():
            if name in state:
                setattr(self, name, state[name])

    def state_write(self, f):
        vaex.utils.write_json_or_yaml(f, self.state_get())

    def state_load(self, f):
        state = vaex.utils.read_json_or_yaml(f)
        self.state_set(state)





import os
import warnings
import numpy as np

import vaex
import vaex.dataset
from vaex.utils import InnerNamespace

from .pipeline import Pipeline
from .transformations import PCA, StandardScaler, MinMaxScaler


def iris():
    dirname = os.path.dirname(__file__)
    return vaex.open(os.path.join(dirname, 'iris.hdf5'))


def pca(self, n_components=2, features=None, progress=False):
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    pca = PCA(n_components=n_components, features=features)
    pca.fit(self, progress=progress)
    return pca


def standard_scaler(self, features=None, with_mean=True, with_std=True):
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    standard_scaler = StandardScaler(features=features, with_mean=with_mean, with_std=with_std)
    standard_scaler.fit(self)
    return standard_scaler


def minmax_scaler(self, features=None, feature_range=[0, 1]):
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    minmax_scaler = MinMaxScaler(features=features, feature_range=feature_range)
    minmax_scaler.fit(self)
    return minmax_scaler


def xgboost_model(self, label, num_round, features=None, copy=False, **param):
    from .xgboost import XGBModel
    dataset = self
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    xg = XGBModel(num_round=num_round, features=features, **param)
    xg.fit(dataset, label, copy=copy)
    return xg


def state_transfer(self):
    from .transformations import StateTransfer
    return StateTransfer(state=self.state_get())


def one_hot_encoding(self, expression, prefix=None, numeric=True, zero=None, one=None):
    prefix = prefix or expression
    uniques = self.unique(expression)
    uniques = np.sort(uniques)  # gives consistent results

    if numeric:
        if zero is None:
            zero = 0
        if one is None:
            one = 1
    else:
        if zero is None:
            zero = False
        if one is None:
            one = True
    # add a virtual column for each unique value
    for value in uniques:
        column_name = prefix + "_" + str(value)
        self.add_virtual_column(column_name, 'where({expression} == {value}, {one}, {zero})'.format(
            expression=expression, value=repr(value), one=one, zero=zero))


def train_test_split(self, test_size=0.2, strings=True, virtual=True):
    """Will split the dataset in train and test part.

    TODO:
     * see to_copy, we now copy virtual columns
     * maybe respect initial active_range
    """
    warnings.warn('make sure the dataset is shuffled')
    initial = None
    try:
        assert self.filtered is False, 'filtered dataset not supported yet'
        # full_length = len(self)
        self = self.trim()
        initial = self.get_active_range()
        self.set_active_fraction(test_size)
        test = self.trim()
        __, end = self.get_active_range()
        self.set_active_range(end, self.length_original())
        train = self.trim()
    finally:
        if initial is not None:
            self.set_active_range(*initial)
    return train, test


def add_namespace():
    vaex.dataset.Dataset.ml = InnerNamespace({})
    # try:
    #     from . import generated
    # except ImportError:
    #     print("importing generated code failed")
    vaex.dataset.Dataset.ml._add(train_test_split=train_test_split)

    def to_xgboost_dmatrix(self, label, features=None, selection=None, blocksize=1000 * 1000):
        """

        label: ndarray containing the labels
        """
        from . import xgboost
        return xgboost.VaexDMatrix(self, label, features=features, selection=selection, blocksize=blocksize)

    vaex.dataset.Dataset.ml._add(to_xgboost_dmatrix=to_xgboost_dmatrix, xgboost_model=xgboost_model,
                                 state_transfer=state_transfer,
                                 one_hot_encoding=one_hot_encoding,
                                 pca=pca,
                                 standard_scaler=standard_scaler,
                                 minmax_scaler=minmax_scaler)
    del to_xgboost_dmatrix

    # named_objects.update({ep.name: ep.load()})



import sys
import os
import re

import traitlets

from . import state
from . import transformations
from . import generate

#registry = {}
registry = []
def register(cls):
    registry.append(cls)
    #registry[(cls.__module__, cls.__name__)] = cls
    return cls

def camel_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def generate_vaex_ml():
    from jinja2 import Template
    template_method = Template("""

import {{ module }}
import traitlets

def {{ method_name }}(self, {{ signature }}):
    obj = {{ module }}.{{ class_name }}({{ args }})
    obj.fit(self{{ extra_fit }})
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add({{ method_name }}={{ method_name }})

def __init__(self, {{ full_signature }}):
    super({{ module }}.{{ class_name }}, self).__init__({{ full_args }})

{{ module }}.{{ class_name }}.__init__ = __init__
del __init__
    """)
    filename = os.path.join(os.path.dirname(__file__), "generated.py")
    with open(filename, "w") as f:
        for cls in generate.registry:
            traits = {key:value for key, value in cls.class_traits().items() if 'output' not in value.metadata}
            traits_nodefault = {key:value for key, value in traits.items()
                                if value.default_value is traitlets.Undefined}
            traits_default = {key:value for key, value in traits.items()
                                if key not in traits_nodefault}
            signature_list = ["{name}".format(name=name, value=value.default_value)
                for name, value in traits_nodefault.items()]
            signature_list.extend(["{name}={value!r}".format(name=name, value=value.default_value)
                for name, value in traits_default.items()])
            signature = ", ".join(signature_list)
            args = ", ".join(["{name}={name}".format(name=name, value=value)
                for name, value in traits.items()])

            signature_list = ["{name}={value!r}".format(name=name, value=value.default_value)
                for name, value in traits.items()]
            full_signature = ", ".join(signature_list)
            full_args = args

            method_name = camel_to_underscore(cls.__name__)
            module = cls.__module__
            class_name = cls.__name__

            kwargs = dict(locals())
            code = template_method.render(**kwargs)
            print(code, file=f)

def main(argv=sys.argv):
    generate_vaex_ml()
if __name__ == '__main__':
    main()


from setuptools import setup
import sys, os, imp

import sys, os, imp
from setuptools import Extension
from setuptools.command.develop import develop


dirname = os.path.dirname(__file__)
path_version = os.path.join(dirname, "vaex/ml/_version.py")
version = imp.load_source('version', path_version)

name        = 'vaex'
author      = "Maarten A. Breddels"
author_email= "maartenbreddels@gmail.com"
license     = 'Custom'
version     = version.__version__
url         = 'https://www.github.com/maartenbreddels/vaex'
install_requires_ml = ['vaex-core', 'numba']

class DevelopCmd(develop):
    def run(self):
        super(DevelopCmd, self).run()
        import vaex

        source = os.path.abspath(os.path.join('vaex', 'ml'))
        target = os.path.abspath(os.path.join(os.path.dirname(vaex.__file__), 'ml'))
        if not os.path.exists(target):
            print("*** extra: linking ", source, "->", target)
            os.symlink(source, target)


setup(name=name+'-ml',
    version=version,
    description='Machine learning support for vaex',
    url=url,
    author=author,
    author_email=author_email,
    install_requires=install_requires_ml,
    license=license,
    packages=['vaex.ml'],
    zip_safe=False,
    cmdclass={'develop': DevelopCmd},
    entry_points={'vaex.namespace': ['ml = vaex.ml:add_namespace']}
    )



import os
import ctypes

import vaex
import xgboost.core
import numpy as np
import tempfile
import base64
import vaex.serialize
from . import state
import traitlets

#libpath = os.path.join(os.path.dirname(__file__), "vaex_ml_lib.cpython-35m-x86_64-linux-gnu.so")
#print(libpath)
#lib = ctypes.cdll.LoadLibrary(libpath)
lib = xgboost.core._LIB

class EXPRESSION_RESULT(ctypes.Structure):
     _fields_ = [
		 ("length", ctypes.c_longlong),
		 ("data", ctypes.POINTER(ctypes.c_float))
	 ]

EVALFUNC = ctypes.CFUNCTYPE(ctypes.c_int, (ctypes.c_char_p), ctypes.c_longlong, ctypes.c_longlong, ctypes.POINTER(EXPRESSION_RESULT))

@vaex.serialize.register
class XGBModel(state.HasState):

	features = traitlets.List(traitlets.Unicode())
	num_round = traitlets.CInt()
	param = traitlets.Dict()
	prediction_name = traitlets.Unicode(default_value='xgboost_prediction')


	def __call__(self, *args):
		data2d = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
		dmatrix = xgboost.DMatrix(data2d)
		return self.bst.predict(dmatrix)

	def transform(self, dataset):
		copy = dataset.copy()
		lazy_function = copy.add_function('xgboost_prediction_function', self)
		expression = lazy_function(*self.features)
		copy.add_virtual_column(self.prediction_name, expression, unique=False)
		return copy

	def fit(self, dataset, label, copy=False):
		if copy:
			data = np.vstack([dataset.evaluate(k).astype(np.float64) for k in self.features]).T
			label_data = dataset.evaluate(label)
			dtrain = xgboost.DMatrix(data, label_data)
		else:
			dtrain = VaexDMatrix(dataset, label, features=self.features)
		param = dict(self.param)
		self.bst = xgboost.train(param, dtrain, self.num_round)

	def predict(self, dataset, copy=False):
		if copy:
			data = np.vstack([dataset.evaluate(k) for k in self.features]).T
			dmatrix = xgboost.DMatrix(data)
		else:
			dmatrix = VaexDMatrix(dataset, features=self.features)
		return self.bst.predict(dmatrix)

	def state_get(self):
		filename = tempfile.mktemp()
		self.bst.save_model(filename)
		with open(filename, 'rb') as f:
			data = f.read()
		return dict(tree_state=base64.encodebytes(data).decode('ascii'),
			substate=super(XGBModel, self).state_get())

	def state_set(self, state):
		super(XGBModel, self).state_set(state['substate'])
		data = base64.decodebytes(state['tree_state'].encode('ascii'))
		filename = tempfile.mktemp()
		with open(filename, 'wb') as f:
			f.write(data)
		self.bst = xgboost.Booster(model_file=filename)

class VaexDMatrix(xgboost.DMatrix):
	def __init__(self, ds, label=None, features=None, selection=None, blocksize=10):
		super(VaexDMatrix, self).__init__(None)
		self.ds = ds
		self.features = features or self.ds.get_column_names(virtual=True)
		self.selection = selection
		self.blocksize = blocksize

		self.data_references = {}

		self.c_features = (ctypes.c_char_p * len(self.features))()
		self.c_features[:] = [k.encode() for k in self.features]

		self.c_evaluate = EVALFUNC(self.evaluate)

		self.handle = ctypes.c_void_p()
		rows = int(self.ds.count(selection=selection))
		batch_size = 1*1000*1000
		return_code = lib.create_dmatrix(self.c_features, len(self.features), rows, batch_size, self.c_evaluate, self.blocksize, ctypes.byref(self.handle))
		#lib.test_dmatrix(self.handle, 0)
		if label is not None:
			label = self.ds.evaluate(label)
			self.set_label_npy2d(label)

	def set_label_npy2d(self, label):
		"""Set label of dmatrix

		Parameters
		----------
		label: array like
		    The label information to be set into DMatrix
		    from numpy 2D array
		"""
		self.set_float_info_npy2d('label', label)

	def set_float_info_npy2d(self, field, data):
		"""Set float type property into the DMatrix
		   for numpy 2d array input

		Parameters
		----------
		field: str
		    The field name of the information

		data: numpy array
		    The array of data to be set
		"""
		data = np.array(data, copy=False, dtype=np.float32)
		c_data = data.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
		xgboost.core._check_call(xgboost.core._LIB.XGDMatrixSetFloatInfo(self.handle,
		                                       xgboost.core.c_str(field),
		                                       c_data,
		                                       ctypes.c_uint64(len(data))))

	def evaluate(self, expression, i1, i2, result):
		try:
			expression = expression.decode()
			#print(expression, i1, i2)
			#print(result[0])
			data = self.ds.evaluate(expression, i1, i2) # TODO, support selections
			# keep a reference, since otherwise the memory is freed
			# and we overwrite the last reference, so that will be freed
			self.data_references[expression] = data = data.astype(np.float32) # TODO: think about masking, nan?
			#print(data, type(data), data.dtype)
			ptr = data.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
			#print(ptr, ptr)
			result[0].length = len(data)
			result[0].data = ptr
			#print(result.data, result.length)
			#print(result)
			return 0 #EXPRESSION_RESULT(len(data), ptr)
		except (Exception, e):
			print('error', e)
